from __future__ import annotations

import shutil
import xml.etree.cElementTree as ET
from pathlib import Path

import numpy as np
import trimesh


def generate_temp_urdf(
    mesh: trimesh.Trimesh | str,
    tempdir: str,
    rgba: list[int] = [1, 1, 1, 1],
    col_mesh: trimesh.Trimesh | str = None,
):
    meshes = {"visual": mesh, "collision": col_mesh}
    mesh_pathes = {}
    for vis_col in ["collision", "visual"]:
        mesh_path = Path(tempdir) / f"mesh_{vis_col}.obj"
        if mesh_path.exists():
            mesh_path.unlink()

        if isinstance(mesh, str):
            shutil.copy(Path(meshes["visual"]), mesh_path)
            mesh = trimesh.load(mesh_path)
        if isinstance(mesh, trimesh.Trimesh):
            mesh.export(mesh_path, "obj")
        mesh_pathes[vis_col] = Path(mesh_path)
    mesh_offset = np.zeros(3)  # - mesh.centroid

    ndarray_to_str = lambda x: str(x.round(3)).replace("[", "").replace("]", "")
    offset_str = ndarray_to_str(mesh_offset)
    color_str = ndarray_to_str(np.array(rgba))

    urdf_name = "temp"
    root = ET.Element("robot", name=urdf_name + ".urdf")
    link = ET.SubElement(root, "link", name="mesh")

    visual = ET.SubElement(link, "visual")
    collision = ET.SubElement(link, "collision")
    elements = {"visual": visual, "collision": collision}
    for name in ["visual", "collision"]:
        element = elements[name]
        origin = ET.SubElement(element, "origin", xyz=offset_str, rpy="0 0 0")
        geometry = ET.SubElement(element, "geometry")
        geometry_mesh = ET.SubElement(
            geometry, "mesh", filename=f"mesh_{name}.obj", scale="1 1 1"
        )
    material = ET.SubElement(visual, "material", name="color")
    ET.SubElement(material, "color", rgba=color_str)

    inertial = ET.SubElement(link, "inertial")
    ET.SubElement(inertial, "mass", value="0.5")
    ET.SubElement(
        inertial, "inertia", ixx="0.1", ixy="0", ixz="0", iyy="0.1", iyz="0", izz="0.1"
    )

    tree = ET.ElementTree(root)
    ET.indent(tree, "  ")
    urdf_path = Path(tempdir) / "temp.urdf"
    tree.write(urdf_path, encoding="utf-8", xml_declaration=True)
    return urdf_path


def generate_frame_urdf(tempdir, length=0.05, radius=0.005):
    urdf_name = "frame"
    root = ET.Element("robot", name=urdf_name)
    base_link = ET.SubElement(root, "link", name="base")

    def add_material(visual, color_name, rgba_str):
        material = ET.SubElement(visual, "material", name=color_name)
        ET.SubElement(material, "color", rgba=rgba_str)

    def add_fixed_joint(
        name, parent_name, child_name, xyz_str="0 0 0", rpy_str="0 0 0"
    ):
        joint = ET.SubElement(root, "joint", name=name, type="fixed")
        ET.SubElement(joint, "parent", link=parent_name)
        ET.SubElement(joint, "child", link=child_name)
        ET.SubElement(joint, "origin", rpy=rpy_str, xyz=xyz_str)
        ET.SubElement(joint, "axis", xyz="0 0 0")

    def add_axis_link(
        name,
        color_name,
        rgba_str,
        length=0.05,
        radius=0.005,
    ):
        link = ET.SubElement(root, "link", name=name)
        visual = ET.SubElement(link, "visual")
        geometry = ET.SubElement(visual, "geometry")
        inertial = ET.SubElement(link, "inertial")
        ET.SubElement(geometry, "cylinder", length=f"{length}", radius=f"{radius}")
        ET.SubElement(visual, "origin", xyz=f"0 0 {length/2}", rpy="0 0 0")
        add_material(visual, color_name, rgba_str)
        ET.SubElement(inertial, "mass", value="0.1")
        ET.SubElement(
            inertial,
            "inertia",
            ixx="0.01",
            ixy="0",
            ixz="0",
            iyy="0.01",
            iyz="0",
            izz="0.01",
        )

    visual = ET.SubElement(base_link, "visual")
    geometry = ET.SubElement(visual, "geometry")
    inertial = ET.SubElement(base_link, "inertial")
    ET.SubElement(inertial, "mass", value="0.1")
    ET.SubElement(
        inertial,
        "inertia",
        ixx="0.01",
        ixy="0",
        ixz="0",
        iyy="0.01",
        iyz="0",
        izz="0.01",
    )
    ET.SubElement(geometry, "sphere", radius="0.01")
    ET.SubElement(visual, "origin", xyz="0 0 0", rpy="0 0 0")
    add_material(visual, "white", "1 1 1 1")
    add_axis_link("x_axis", "red", "1 0 0 1", length=length, radius=radius)
    add_axis_link("y_axis", "green", "0 1 0 1", length=length, radius=radius)
    add_axis_link("z_axis", "blue", "0 0 1 1", length=length, radius=radius)
    add_fixed_joint(
        "x_axis_joint",
        "base",
        "x_axis",
        rpy_str="0 1.5708 0",
    )
    add_fixed_joint(
        "y_axis_joint",
        "base",
        "y_axis",
        rpy_str="-1.5708 0 0",
    )
    add_fixed_joint(
        "z_axis_joint",
        "base",
        "z_axis",
        rpy_str="0 0 0",
    )

    urdf_path = Path(tempdir) / "frame.urdf"
    tree = ET.ElementTree(root)
    ET.indent(tree, "  ")
    tree.write(urdf_path, encoding="utf-8", xml_declaration=True)
    return urdf_path



import numpy as np
import pybullet as p
from attrs import Factory, define, field
from ezpose import SE3


@define
class ShapeBase:
    rgba: tuple = (1, 1, 1, 1)
    ghost: bool = True
    offset: SE3 = Factory(SE3)

    def get_viz_query(self) -> dict:
        return dict(
            visualFramePosition=self.offset.trans,
            visualFrameOrientation=self.offset.rot.as_quat(),
            rgbaColor=self.rgba,
        )

    def get_col_query(self):
        return dict(
            collisionFramePosition=self.offset.trans,
            collisionFrameOrientation=self.offset.rot.as_quat(),
        )

    def __eq__(self, other):
        d1 = {**self.get_viz_query(), **self.get_col_query()}
        d2 = {**other.get_viz_query(), **other.get_col_query()}
        return all(
            [
                d1[k] == d2[k]
                if not isinstance(d1[k], np.ndarray)
                else np.allclose(d1[k], d2[k])
                for k in d1.keys()
            ]
        )

    def __hash__(self):
        d = {**self.get_viz_query(), **self.get_col_query()}
        s = str({k: d[k] for k in d.keys()})
        return s.__hash__()


@define(frozen=True, kw_only=True)
class SphereShape(ShapeBase):
    radius: float

    def get_viz_query(self):
        query = super().get_viz_query()
        query.update(shapeType=p.GEOM_SPHERE, radius=self.radius)
        return query

    def get_col_query(self):
        query = super().get_col_query()
        query.update(shapeType=p.GEOM_SPHERE, radius=self.radius)
        return query

    def __eq__(self, other):
        return super().__eq__(other)

    def __hash__(self):
        return super().__hash__()


@define(frozen=True, kw_only=True)
class CylinderShape(ShapeBase):
    radius: float
    length: float

    def get_viz_query(self):
        query = super().get_viz_query()
        query.update(
            shapeType=p.GEOM_CYLINDER,
            radius=self.radius,
            length=self.length,
        )
        return query

    def get_col_query(self):
        query = super().get_col_query()
        query.update(
            shapeType=p.GEOM_CYLINDER,
            radius=self.radius,
            height=self.length,
        )
        return query

    def __eq__(self, other):
        return super().__eq__(other)

    def __hash__(self):
        return super().__hash__()


@define(frozen=True, kw_only=True)
class BoxShape(ShapeBase):
    half_extents: tuple[float] = field(converter=tuple)

    def get_viz_query(self):
        query = super().get_viz_query()
        query.update(
            shapeType=p.GEOM_BOX,
            halfExtents=self.half_extents,
        )
        return query

    def get_col_query(self):
        query = super().get_col_query()
        query.update(
            shapeType=p.GEOM_BOX,
            halfExtents=self.half_extents,
        )
        return query

    def __eq__(self, other):
        return super().__eq__(other)

    def __hash__(self):
        return super().__hash__()


@define(frozen=True, kw_only=True)
class MeshShape(ShapeBase):
    visual_mesh_path: str = field(converter=str)
    col_mesh_path: str = field(converter=str)
    scale: float

    def get_viz_query(self):
        query = super().get_viz_query()
        query.update(
            shapeType=p.GEOM_MESH,
            fileName=self.visual_mesh_path,
            meshScale=np.ones(3) * self.scale,
        )
        return query

    def get_col_query(self):
        query = super().get_col_query()
        query.update(
            shapeType=p.GEOM_MESH,
            fileName=self.col_mesh_path,
            meshScale=np.ones(3) * self.scale,
        )
        return query

    def __eq__(self, other):
        return super().__eq__(other)

    def __hash__(self):
        return super().__hash__()



import pybullet as p
import ezbullet as eb
from ezpose import SE3, SO3
import trimesh
import numpy as np

world = eb.World(gui=True, z_gravity=0.)
ground_mesh = trimesh.creation.box(extents=[10, 10, 0.1])
ground = eb.URDF.from_trimesh("ground", world, ground_mesh, ground_mesh, fixed=True)
ground.set_pose(SE3(trans=[0,0,-0.05]))

def cb():
    tires: list[eb.Body] = [body for name, body in world.bodies.items() if "tire" in name]
    for tire in tires:
        pose = tire.get_pose()
        rotvec = pose.rot.as_rotvec()
        rotvec[0] = 0.
        rotvec[2] = 0.
        pose.rot = SO3.from_rotvec(rotvec)
        pose.trans[1] = 0. # xz constr
        tire.set_pose(pose)

world.set_step_callback(cb)
body1 = eb.Box.create("tire1", world, [0.05, 0.05, 0.05])
body2 = eb.Box.create("tire2", world, [0.05, 0.05, 0.05])


body1.set_pose(SE3(trans=[0,0,0.2]))
body2.set_pose(SE3(trans=[0.1,0,0.3]))

while True:
    world.step()


from __future__ import annotations

from contextlib import contextmanager
from pathlib import Path

import numpy as np
import pybullet as p
import trimesh
from attrs import define, field
from ezpose import SE3, SO3
from numpy.typing import ArrayLike

from .data import JointInfo, JointState
from .utils import generate_frame_urdf, generate_temp_urdf
from .world import Body, World


@define(repr=False)
class URDF(Body):
    path: str
    fixed: bool
    scale: float
    dof: int
    joint_info: list[JointInfo]
    movable_joints: np.ndarray
    pos_ctrl_gain_p: list[float]
    pos_ctrl_gain_d: list[float]
    max_torque: list[float]
    T_com: SE3 = field(init=False)

    def __attrs_post_init__(self):
        self.T_com = super().get_pose()
        self.set_joint_angles(self.neutral)

    @classmethod
    def create(
        cls,
        name,
        world: World,
        path: Path | str,
        fixed: bool = True,
        scale: float = 1.0,
        ghost: bool = False,
        position_gain: float = 0.1,
        velocity_gain: float = 1.0,
    ):
        if name in world.bodies_and_ghosts:
            print("Body name already exists.")
            return world.bodies_and_ghosts[name]
            # world.remove_body(name)

        # flags = p.URDF_USE_INERTIA_FROM_FILE if mass is not None else 0
        uid = world.loadURDF(
            fileName=str(path),
            useFixedBase=fixed,
            globalScaling=scale,
            # flags=flags, # TODO
        )
        dof = world.getNumJoints(uid)

        joint_info = []
        movable_joints = []
        for i in range(dof):
            info = JointInfo(*world.getJointInfo(uid, i))
            joint_info.append(info)
            if info.movable:
                movable_joints.append(i)
        movable_joints = np.array(movable_joints)
        position_gain = [position_gain] * len(movable_joints)
        pos_ctrl_gain_d = [velocity_gain] * len(movable_joints)

        max_torque = [
            info.joint_max_force
            for info in joint_info
            if info.joint_index in movable_joints
        ]

        mass = None
        body = cls(
            world=world,
            uid=uid,
            name=name,
            mass=mass,
            ghost=ghost,
            path=path,
            fixed=fixed,
            scale=scale,
            dof=dof,
            joint_info=joint_info,
            movable_joints=movable_joints,
            pos_ctrl_gain_p=position_gain,
            pos_ctrl_gain_d=pos_ctrl_gain_d,
            max_torque=max_torque,
        )

        if ghost:
            # TODO -> setCollisionFilterGroupMask
            # TODO: Do something here if collision behavior should be changed
            world.ghosts[name] = body
            body.disable_collision()

        else:
            world.bodies[name] = body
        return body

    @classmethod
    def from_trimesh(
        cls,
        name: str,
        world: World,
        visual_mesh: trimesh.Trimesh,
        col_mesh: trimesh.Trimesh,
        fixed: bool,
        rgba=[1, 1, 1, 1],
    ):
        import tempfile

        with tempfile.TemporaryDirectory() as tempdir:
            urdf_path = generate_temp_urdf(
                visual_mesh, tempdir, rgba, col_mesh=col_mesh
            )
            obj = cls.create(name, world, urdf_path, fixed=fixed, scale=1.0)
        return obj

    @property
    def lb(self):
        return np.array(
            [joint.joint_lower_limit for joint in self.joint_info if joint.movable]
        )

    @property
    def ub(self):
        return np.array(
            [joint.joint_upper_limit for joint in self.joint_info if joint.movable]
        )

    @property
    def neutral(self):
        return (self.lb + self.ub) / 2

    @contextmanager
    def set_joint_angles_context(self, q):
        joints_temp = self.get_joint_angles()
        self.set_joint_angles(q)
        yield
        self.set_joint_angles(joints_temp)

    def set_pose(self, pose):
        """This is because, super().set_pose is setting a pose of base inertial frame.
        This differs from initial load state of the URDF."""
        super().set_pose(pose @ self.T_com)

    def get_pose(self):
        return super().get_pose() @ self.T_com.inv()

    def get_joint_states(self):
        return [
            JointState(*s)
            for s in self.world.getJointStates(self.uid, self.movable_joints)
        ]

    def get_joint_angles(self):
        return np.array([s.pos for s in self.get_joint_states()])

    def get_joint_velocities(self):
        return np.array([s.vel for s in self.get_joint_states()])

    def set_joint_angle(self, i, angle):
        self.world.resetJointState(self.uid, jointIndex=i, targetValue=angle)

    def set_joint_angles(self, angles):
        assert len(angles) == len(
            self.movable_joints
        ), f"num_angle is not matched: {len(angles)} vs {len(self.movable_joints)}"
        for i, angle in zip(self.movable_joints, angles):
            self.set_joint_angle(i, angle)

    def get_joint_frame_pose(self, joint_idx):
        assert len(self.joint_info) > joint_idx
        parent_link_idx = joint_idx - 1
        parent_link_pose = self.get_link_pose(parent_link_idx)
        pos = self.joint_info[joint_idx].parent_frame_pos
        quat = self.joint_info[joint_idx].parent_frame_orn
        rel_pose = SE3(SO3.from_quat(quat), pos)
        return parent_link_pose @ rel_pose

    def get_link_idx_by_name(self, link_name: str):
        idx = [j.joint_index for j in self.joint_info if j.link_name == link_name]
        if len(idx) == 0:
            return None
        return idx[0]

    def get_link_pose(self, link_idx: int | str, frame="com"):
        if isinstance(link_idx, str):
            link_idx = self.get_link_idx_by_name(link_idx)
        assert len(self.joint_info) > link_idx
        if link_idx == -1:
            return super().get_pose()
        if frame == "com":
            pos, xyzw = self.world.getLinkState(self.uid, link_idx)[:2]
        elif frame == "urdf":
            pos, xyzw = self.world.getLinkState(self.uid, link_idx)[4:6]
        return SE3(SO3.from_quat(xyzw), pos)

    def get_aabb(self, link_idx: int | str = "all"):
        """if link_idx == "all", it computes the overall boundingbox"""
        if link_idx == "all":
            lb, ub = np.full(3, np.inf), np.full(3, -np.inf)
            link_indices = [-1] + [i for i in range(self.dof)]
            for idx in link_indices:
                lower, upper = super().get_aabb(idx)
                lb, ub = np.minimum(lb, lower), np.maximum(ub, upper)
            return lb, ub
        else:
            return super().get_aabb(link_idx)

    def forward_kinematics(self, q: ArrayLike, link_idx: int):
        with self.set_joint_angles_context(q):
            pose = self.get_link_pose(link_idx, frame="urdf")
        return pose

    def inverse_kinematics(
        self,
        target_pose: SE3,
        link_idx: int,
        validate=True,
        max_iter=10,
        pos_tol=1e-3,
        start_with_neutral=True,
        verbose=False,
    ):
        solved = False
        if start_with_neutral:
            q = self.neutral
        else:
            q = self.get_joint_angles()
        if not isinstance(link_idx, int):
            link_idx = self.get_link_idx_by_name(link_idx)

        with self.set_joint_angles_context(q):
            for _ in range(max_iter):
                ik_sol = self.world.calculateInverseKinematics(
                    bodyIndex=self.uid,
                    endEffectorLinkIndex=link_idx,
                    targetPosition=target_pose.trans,
                    targetOrientation=target_pose.rot.as_quat(),
                )
                # update initial joint angles to ik solution
                self.set_joint_angles(ik_sol)
                if not validate:
                    solved = True
                    break

                pose_sol = self.forward_kinematics(ik_sol, link_idx)
                trans_err = np.linalg.norm(pose_sol.trans - target_pose.trans)
                is_in_bound = np.all(ik_sol >= self.lb) and np.all(ik_sol <= self.ub)
                is_near_target = trans_err < pos_tol
                if is_near_target and is_in_bound:
                    solved = True
                    break

        if not solved:
            if verbose:
                print(f"ik not solved. pos error = {trans_err}")
            return None
        else:
            if verbose:
                print(f"ik solved. pos error = {trans_err}")
            return np.array(ik_sol)  # if solved else None

    def set_ctrl_target_joint_angles(self, q, max_torque=None):
        assert len(q) == len(self.movable_joints)
        if max_torque is None:
            max_torque = self.max_torque
        elif isinstance(max_torque, float):
            max_torque = [max_torque] * len(self.movable_joints)

        self.world.setJointMotorControlArray(
            self.uid,
            jointIndices=self.movable_joints,
            controlMode=p.POSITION_CONTROL,
            targetPositions=q,
            forces=max_torque,
            positionGains=self.pos_ctrl_gain_p,
            velocityGains=self.pos_ctrl_gain_d,
        )

    def set_ctrl_target_joint_velocities(self, qdot, max_torque=None):
        assert len(qdot) == len(self.movable_joints)
        if max_torque is None:
            max_torque = self.max_torque
        elif isinstance(max_torque, float):
            max_torque = [max_torque] * len(self.movable_joints)

        self.world.setJointMotorControlArray(
            self.uid,
            jointIndices=self.movable_joints,
            controlMode=p.VELOCITY_CONTROL,
            targetVelocities=qdot,
            forces=self.max_torque,
            # positionGains=self.pos_ctrl_gain_p,
            # velocityGains=self.pos_ctrl_gain_d,
        )

    def get_jacobian(self, q, link_idx, local_position=[0, 0, 0]):
        jac_trans, jac_rot = self.world.calculateJacobian(
            bodyUniqueId=self.uid,
            linkIndex=link_idx,
            localPosition=local_position,
            objPositions=q.tolist(),
            objVelocities=np.zeros_like(q).tolist(),
            objAccelerations=np.zeros_like(q).tolist(),
        )
        return np.vstack([jac_trans, jac_rot])


class Frame(URDF):
    @classmethod
    def create(cls, name: str, world: World, length=0.05, radius=0.005):
        import tempfile

        with tempfile.TemporaryDirectory() as tempdir:
            urdf_path = generate_frame_urdf(
                tempdir, length, radius
            )  # TODO:add mass, inertial information
            # Note: "frame.urdf" has no collision shape
            frame = super().create(name, world, urdf_path, fixed=True, ghost=True)
        return frame



import pytest

import numpy as np
from ezpose import SE3, SO3
import ezbullet as eb
import trimesh
import pybullet as p


@pytest.fixture
def world():
    world = eb.World(gui=True)
    ground_mesh = trimesh.creation.box(extents=[10, 10, 0.1])
    ground = eb.URDF.from_trimesh(
        "ground", world, ground_mesh, ground_mesh, fixed=True
    )
    box = eb.Box.create("box", world, [0.05, 0.05, 0.05])
    return world


def test_transform(world: eb.World):
    box = world.bodies["box"]
    box.set_pose(SE3(rot=SO3.random(), trans=[0, 0, 0.1]))


# def test_constraint():
#     name = "box"
#     body = eb.Box.create(name, world, [0.05, 0.05, 0.05])
#     dummy_link_x = eb.Body.create_empty_body(f"{name}_dummy_link_x", world)
#     dummy_link_z = eb.Body.create_empty_body(f"{name}_dummy_link_z", world)
#     x_constr = p.createConstraint(
#         parentBodyUniqueId=-1,
#         parentLinkIndex=-1,
#         childBodyUniqueId=body.uid,
#         childLinkIndex=-1,
#         jointType=p.JOINT_PRISMATIC,
#         jointAxis=[1, 0, 0],  # x축 이동 허용
#         parentFramePosition=[0, 0, 0],
#         childFramePosition=[0, 0, 0]
#     )
#     z_constr = p.createConstraint(
#         parentBodyUniqueId=dummy_link_x.uid,
#         parentLinkIndex=-1,
#         childBodyUniqueId=dummy_link_z.uid,
#         childLinkIndex=-1,
#         jointType=p.JOINT_PRISMATIC,
#         jointAxis=[0, 0, 1],  # z축 이동 허용
#         parentFramePosition=[0, 0, 0],
#         childFramePosition=[0, 0, 0]
#     )
#     world.create_constraint(f"{name}_x_constr", x_constr)
#     world.create_constraint(f"{name}_z_constr", z_constr)
#     world.change_constraint(f"{name}_x_constr", maxForce=100)
#     world.change_constraint(f"{name}_z_constr", maxForce=100)



import pybullet as p
from attrs import define, field

"""Information Structure used in Pybullet
"""
conv_str = lambda x: x.decode("ascii") if isinstance(x, bytes) else x


@define
class JointState:
    pos: float
    vel: float
    wrench: tuple[float]  # reaction force
    torque: float  # applied motor torque


@define
class JointInfo:
    joint_index: int
    joint_name: str = field(converter=conv_str)
    joint_type: int
    q_index: int
    u_index: int
    flags: int
    joint_damping: float
    joint_friction: float
    joint_lower_limit: float
    joint_upper_limit: float
    joint_max_force: float
    joint_max_velocity: float
    link_name: str = field(converter=conv_str)
    joint_axis: tuple[float]
    parent_frame_pos: tuple[float]
    parent_frame_orn: tuple[float]
    parent_index: int

    @property
    def movable(self):
        return self.joint_type != p.JOINT_FIXED


@define
class DistanceInfo:
    contact_flag: bool
    bodyA: int
    bodyB: int
    linkA: int
    linkB: int
    position_on_A: tuple[float]
    position_on_B: tuple[float]
    contact_normal_on_B: tuple[float]
    contact_distnace: float
    normal_force: float
    lateral_frictionA: float
    lateral_friction_dirA: tuple[float]
    lateral_frictionB: float
    lateral_friction_dirB: tuple[float]


@define
class ContactInfo:
    contactFlag: int  # nothing
    bodyA: int
    bodyB: int
    link_indexA: int
    link_indexB: int
    position_on_A: tuple[float]
    position_on_B: tuple[float]
    contact_normal_on_B: tuple[float]
    contact_distance: float
    normal_force: float
    lateral_friction1: float
    lateral_friction_dir1: tuple[float]
    lateral_friction2: float
    lateral_friction_dir2: tuple[float]


@define
class DynamicsInfo:
    mass: float
    lateral_friction: float
    local_inertial_diagonal: tuple[float]
    local_inertial_pos: tuple[float]
    local_inertial_orn: tuple[float]
    restitution: float
    rolling_friction: float
    spinning_friction: float
    contact_damping: float
    contact_stiffness: float
    body_type: int
    collision_margin: float


@define
class ConstraintInfo:
    parent_body: int
    parent_link: int
    child_body: int
    child_link: int
    constr_type: int
    joint_axis: tuple[float]
    parent_frame_pos: tuple[float] = None
    parent_frame_orn: tuple[float] = None
    child_frame_pos: tuple[float] = None
    child_frame_orn: tuple[float] = None

    def to_dict(self):
        res = dict(
            parentBodyUniqueId=self.parent_body,
            parentLinkIndex=self.parent_link,
            childBodyUniqueId=self.child_body,
            childLinkIndex=self.child_link,
            jointType=self.constr_type,
            jointAxis=self.joint_axis,
            parentFramePosition=self.parent_frame_pos,
            parentFrameOrientation=self.parent_frame_orn,
            childFramePosition=self.child_frame_pos,
            childFrameOrientation=self.child_frame_orn,
        )
        return {k: v for k, v in res.items() if v is not None}



from __future__ import annotations

import time
from contextlib import contextmanager

import numpy as np
import pybullet as p
from attrs import define
from ezpose import SE3, SO3
from pybullet_utils.bullet_client import BulletClient
from typing import Callable

from .data import ConstraintInfo, ContactInfo, DistanceInfo, DynamicsInfo
from .shape import ShapeBase


class World(BulletClient):
    worlds: dict[str, World] = dict()
    gui_world_exists = False

    @classmethod
    def __new__(cls, *args, **kwargs):
        """This prevents to create two visualizers"""
        gui = True
        if len(args) > 1:
            gui = args[1]
        elif "gui" in kwargs:
            gui = kwargs["gui"]

        if "gui" in cls.worlds and gui:
            print("You can't create two visualizers")
            print("Load the existing world")
            return cls.worlds["gui"]
        else:
            world = super().__new__(cls)

        if gui:
            # print('create gui world')
            cls.worlds["gui"] = world
        else:
            # print('create no gui world')
            if "no_gui" not in cls.worlds:
                cls.worlds["no_gui"] = []
            cls.worlds["no_gui"].append(world)
        return world

    def __init__(
        self,
        gui=True,
        z_gravity=-9.81,
        dt=0.005,
        solver_iter=150,
        realtime=True,
        realtime_factor=0.5,
        background_color=None,
    ):
        if hasattr(self, "_init"):
            return  # preventing multiple initialization

        self.gui = gui
        self.dt = dt
        self.solver_iter = solver_iter
        self.realtime = realtime
        self.realtime_factor = realtime_factor
        self.gravity = z_gravity

        options = ""
        if background_color is not None:
            background_color = np.array(background_color).astype(np.float64)
            options = f"--background_color_red={background_color[0]} \
                        --background_color_green={background_color[1]} \
                        --background_color_blue={background_color[2]}"

        if gui:
            if self.gui_world_exists:
                return
            else:
                self.gui_world_exists = True
        connection_mode = p.GUI if gui else p.DIRECT

        super().__init__(connection_mode=connection_mode, options=options)
        time.sleep(1.0)

        if self.gui:
            self.pause_button_uid = p.addUserDebugParameter("turn off loop", 1, 0, 1)

        self.reset()
        self.watch_workspace()
        self._init = True
        self._step_cb = None

    @contextmanager
    def no_rendering(self):
        self.configureDebugVisualizer(p.COV_ENABLE_RENDERING, 0)
        yield
        self.configureDebugVisualizer(p.COV_ENABLE_RENDERING, 1)

    def set_step_callback(self, callback:Callable):
        self._step_cb = callback

    def watch_workspace(
        self, target_pos=[0, 0, 0], distance=1.0, cam_yaw=45, cam_pitch=-35
    ):
        self.resetDebugVisualizerCamera(
            cameraDistance=distance,
            cameraYaw=cam_yaw,
            cameraPitch=cam_pitch,
            cameraTargetPosition=target_pos,
        )

    @property
    def bodies_and_ghosts(self):
        return {**self.bodies, **self.ghosts}

    def get_body(self, name):
        if name in self.bodies:
            return self.bodies[name]
        return None

    def set_gravity(self, force_z=-9.81):
        self.setGravity(0, 0, force_z)

    def reset(self):
        self.resetSimulation()
        self.setPhysicsEngineParameter(
            fixedTimeStep=self.dt, numSolverIterations=self.solver_iter
        )
        self.t = 0.0
        self.bodies: dict[str, Body] = dict()
        self.shapes: dict[str, ShapeBase] = dict()
        self.ghosts: dict[str, Body] = dict()
        self.debug_items: dict[str, int] = dict()  # not wrapped by class
        self.constr: dict[str, int] = dict()  # not wrapped by class
        self.set_gravity(self.gravity)

    def step(self, no_dynamics=False):
        if no_dynamics:
            self.performCollisionDetection()
        else:
            self.stepSimulation()
        if self._step_cb is not None:
            self._step_cb()
        
        # add delay for realtime visualization
        if self.gui and self.realtime:
            time.sleep(self.dt * self.realtime_factor)
        self.t += self.dt

    def show(self):
        """Start infinite loop to visualize for macos"""
        num_quit = p.readUserDebugParameter(self.pause_button_uid)

        polling_rate = 100
        while True:
            self.step(no_dynamics=True)
            time.sleep(1 / polling_rate)
            quit = p.readUserDebugParameter(self.pause_button_uid)
            if quit >= num_quit + 1:
                break

    def get_shape_id(self, shape: ShapeBase):
        if shape in self.shapes:
            return self.shapes[shape]

        viz_id = self.createVisualShape(**shape.get_viz_query())
        col_id = (
            -1 if shape.ghost else self.createCollisionShape(**shape.get_col_query())
        )

        self.shapes[shape] = (viz_id, col_id)
        return self.shapes[shape]

    def remove_body(self, body: str | Body | BodyContainer):
        if isinstance(body, str):
            body = self.bodies_and_ghosts[body]

        def _remove_body(body: str | Body | BodyContainer, body_dict: dict[str, Body]):
            if isinstance(body, BodyContainer):
                for b in body.bodies:
                    self.removeBody(b.uid)  # delete all contained bodies
            elif isinstance(body, Body):
                self.removeBody(body.uid)
            del body_dict[body.name]

        if body.ghost:
            _remove_body(body, self.ghosts)
        else:
            _remove_body(body, self.bodies)

    def remove_all_debugitems(self):
        self.removeAllUserDebugItems()

    def remove_debug_item(self, name):
        if name in self.debug_items:
            self.removeUserDebugItem(self.debug_items[name])
            del self.debug_items[name]

    def add_debug_line(self, p_from, p_to, color=[1, 0, 0], name=None):
        """if name is not None, debug uid is tracked in self.debug_items"""
        uid = self.addUserDebugLine(
            lineFromXYZ=p_from, lineToXYZ=p_to, lineColorRGB=color
        )

        if name is not None:
            if name in self.debug_items:
                self.remove_debug_item(name)
            self.debug_items[name] = uid

    def add_debug_pcd(self, points, color=[1, 0, 0], size=5, name=None):
        """if name is not None, debug uid is tracked in self.debug_items"""
        if not isinstance(color, np.ndarray):
            color = np.array(color)
        if len(color.shape) == 1:
            num_points = points.shape[0]
            color = color[None, ...].repeat(num_points, axis=0)
        uid = self.addUserDebugPoints(
            pointPositions=points, pointColorsRGB=color, pointSize=size
        )

        if name is not None:
            if name in self.debug_items:
                self.remove_debug_item(name)
            self.debug_items[name] = uid
        return uid

    def draw_bounding_box(self, length, center=np.zeros(3)):
        vertices = np.array(list(np.ndindex((2, 2, 2))))
        edges = np.array(
            [
                (0, 1),
                (0, 2),
                (0, 4),
                (1, 3),
                (1, 5),
                (2, 3),
                (2, 6),
                (3, 7),
                (4, 5),
                (4, 6),
                (5, 7),
                (6, 7),
            ]
        )
        point_from = (vertices[edges[:, 0]] - 0.5) * length + center
        point_to = (vertices[edges[:, 1]] - 0.5) * length + center

        for p1, p2 in zip(point_from, point_to):
            self.add_debug_line(p1, p2, color=[0.5, 0.5, 0.5])

    def draw_frustum(self, pose_offset: SE3 = None, w=0.1, h=0.07, d=0.05):
        if pose_offset is None:
            pose_offset = SE3()

        w2, h2 = w * 1.5, h * 1.5
        # Define vertices of the frustum (top and bottom faces)
        top_vertices = np.array(
            [
                [-w / 2, -h / 2, 0],
                [w / 2, -h / 2, 0],
                [w / 2, h / 2, 0],
                [-w / 2, h / 2, 0],
            ]
        )

        bottom_vertices = np.array(
            [
                [-w2 / 2, -h2 / 2, d],
                [w2 / 2, -h2 / 2, d],
                [w2 / 2, h2 / 2, d],
                [-w2 / 2, h2 / 2, d],
            ]
        )

        # Combine the top and bottom vertices
        vertices = np.vstack([top_vertices, bottom_vertices])

        # Define the edges connecting the vertices (top, bottom, and sides)
        edges = np.array(
            [
                (0, 1),
                (1, 2),
                (2, 3),
                (3, 0),  # top face edges
                (4, 5),
                (5, 6),
                (6, 7),
                (7, 4),  # bottom face edges
                (0, 4),
                (1, 5),
                (2, 6),
                (3, 7),  # side edges connecting top and bottom
            ]
        )

        # Offset the vertices by the center position
        # vertices += center
        vertices = pose_offset.apply(vertices)

        # Draw the edges as lines
        for edge in edges:
            p1, p2 = vertices[edge[0]], vertices[edge[1]]
            self.add_debug_line(p1, p2, color=[1, 0, 0])

    def save_state(self, return_uid=False):
        if return_uid:
            return self.saveState()
        else:
            self._state_uid = self.saveState()

    def restore_state(self, state_uid=None):
        if state_uid is None:
            state_uid = self._state_uid
        return self.restoreState(state_uid)

    def get_distance_info(
        self,
        body1: Body,
        body2: Body,
        link1: int = None,
        link2: int = None,
        tol: float = 0.0,
    ):
        kwargs = dict()
        kwargs["bodyA"] = body1.uid
        kwargs["bodyB"] = body2.uid
        if link1 is not None:
            kwargs["linkIndexA"] = link1
        if link2 is not None:
            kwargs["linkIndexB"] = link2
        kwargs["distance"] = tol
        results = self.getClosestPoints(**kwargs)
        return [DistanceInfo(*info) for info in results]

    def get_contact_info(
        self,
        body1: Body,
        body2: Body = None,
        link1: int = None,
        link2: int = None,
    ):
        """world.step(no_dynamics=True) should be called before using"""
        kwargs = dict()
        kwargs["bodyA"] = body1.uid
        if body2 is not None:
            kwargs["bodyB"] = body2.uid
        if link1 is not None:
            kwargs["linkIndexA"] = link1
        if link2 is not None:
            kwargs["linkIndexB"] = link2
        results = self.getContactPoints(**kwargs)
        return [ContactInfo(*info) for info in results]

    def wait_to_stablize(
        self, wait_time=0.1, timeout=5.0, tol=0.01, polling_period=100
    ):
        for i in range(int(wait_time / self.dt)):
            self.step()

        t = 0.0
        while t < timeout:
            for _ in range(polling_period):
                self.step()
            t += self.dt * polling_period

            chk_all_bodies_rest = all(
                [
                    np.linalg.norm(body.get_velocity()) < tol
                    for body in self.bodies.values()
                ]
            )

            if chk_all_bodies_rest:
                break
        return

    def create_constraint(self, name, constr_info: ConstraintInfo):
        if name in self.constr:
            # constraint exists already
            return
        if constr_info.child_body is None:
            constr_info.child_body = -1

        self.constr[name] = self.createConstraint(**constr_info.to_dict())

    # TODO: make it clear
    def change_constraint(self, name, **kwargs):
        self.changeConstraint(self.constr[name], **kwargs)

    def remove_constraint(self, name):
        if name in self.constr:
            self.removeConstraint(self.constr[name])
            del self.constr[name]

    def apply_external_force(
        self, body: Body, force: np.ndarray, pos: np.ndarray, link: int = -1
    ):
        """Note that this method will only work 
        when explicitly stepping the simulation using stepSimulation"""
        p.applyExternalForce(body.uid, link, force, pos, p.WORLD_FRAME)


@define(repr=False)
class Body:
    world: World
    uid: int
    name: str
    mass: float
    ghost: bool

    def set_pose(self, pose: SE3):
        self.world.resetBasePositionAndOrientation(
            self.uid, pose.trans, pose.rot.as_quat()
        )

    def get_pose(self):
        pos, orn = self.world.getBasePositionAndOrientation(self.uid)
        return SE3(SO3.from_quat(orn), pos)

    def get_velocity(self):
        linear, angular = self.world.getBaseVelocity(self.uid)
        return linear, angular

    def is_collision_with(
        self, other_body: Body, ignore_fixed_base: bool = True, tol=0.0, verbose=False
    ):
        distance_info = self.world.get_distance_info(self, other_body, tol=tol)
        if hasattr(self, "fixed") and self.fixed and ignore_fixed_base:
            # ignore base collisions of fixed bodies
            distance_info = [info for info in distance_info if info.linkA != -1]

        if verbose:
            if len(distance_info) == 0:
                print("no collision")
            else:
                print(f"Body {self.name} collision with {other_body.name}")
        return any(distance_info)

    def is_in_collision(
        self,
        exclude: tuple[Body | BodyContainer] = tuple(),
        tol=0.0,
        ignore_fixed_base=False,
        verbose=False,
    ):
        for other in self.world.bodies.values():
            if other is self or other in exclude:
                continue

            if isinstance(other, BodyContainer):
                for _other in other.bodies:
                    if self is _other:
                        continue
                    if self.is_collision_with(
                        _other,
                        tol=tol,
                        ignore_fixed_base=ignore_fixed_base,
                        verbose=verbose,
                    ):
                        return True
            else:
                if self.is_collision_with(
                    other, tol=tol, ignore_fixed_base=ignore_fixed_base, verbose=verbose
                ):
                    return True
        return False

    def disable_collision(self):
        self.world.setCollisionFilterGroupMask(self.uid, -1, 0, 0)
        if hasattr(self, "joint_info"):
            for link_idx in range(len(self.joint_info)):
                self.world.setCollisionFilterGroupMask(self.uid, link_idx, 0, 0)

    def get_dynamics_info(self, link_idx=-1):
        return DynamicsInfo(*self.world.getDynamicsInfo(self.uid, link_idx))

    def set_dynamics_info(self, link_idx=-1, **kwargs):
        """input_dict should contain key and value of the changeDynamics()
        mass, lateralFriction, restitution, rollingFriction, spinningFriction ..."""
        self.world.changeDynamics(
            bodyUniqueId=self.uid, linkIndex=link_idx, **kwargs
        )

    def get_aabb(self, link_idx=-1):
        lower, upper = self.world.getAABB(self.uid, linkIndex=link_idx)
        lower, upper = np.array(lower), np.array(upper)
        return lower, upper

    def change_color(self, rgba, link_idx=-1):
        self.world.changeVisualShape(
            objectUniqueId=self.uid, linkIndex=link_idx, rgbaColor=rgba
        )

    @classmethod
    def create_empty_body(self, name, world):
        mass = 0.
        uid = p.createMultiBody(
            baseMass=mass,
            baseCollisionShapeIndex=-1,
            baseVisualShapeIndex=-1,
            basePosition=[0., 0., 0.])
        return Body(
            world=world, 
            uid=uid, 
            name=name, 
            mass=mass, 
            ghost=True
        )

    def __repr__(self):
        return f"{self.__class__.__name__}:{self.name}"
    


@define
class BodyContainer:
    """Body container"""

    name: str
    world: World
    bodies: list[Body]
    relative_poses: list[Body]
    ghost: bool

    def __attrs_post_init__(self):
        for body in self.bodies:
            del self.world.bodies[body.name]
        self.world.bodies[self.name] = self

    @classmethod
    def from_bodies(cls, name: str, bodies: list[Body]):
        """The first body will be the reference body"""
        world = bodies[0].world
        rel_poses = [body.get_pose() for body in bodies]
        ref_pose = rel_poses[0]
        rel_poses = [ref_pose.inv() @ pose for pose in rel_poses]
        ghost = all([body.ghost for body in bodies])
        return cls(name, world, bodies, rel_poses, ghost)

    def get_pose(self):
        return self.bodies[0].get_pose()

    def set_pose(self, pose: SE3):
        # self.pose = pose
        poses = [pose @ rel_pose for rel_pose in self.relative_poses]
        for pose, body in zip(poses, self.bodies):
            body.set_pose(pose)

    def get_velocity(self):
        return self.bodies[0].get_velocity()

    def is_in_collision(self, verbose=False):
        for other in self.world.bodies.values():
            if self is other:
                continue

            for body in self.bodies:
                if body.is_in_collision(verbose=verbose):
                    return True
        return False



import numpy as np
import pybullet as p
from attrs import define
from ezpose import SE3, SO3
from numpy.typing import ArrayLike

from .world import World


@define
class CameraIntrinsic:
    width: float
    height: float
    fx: float
    fy: float
    cx: float
    cy: float
    near: float
    far: float

    def get_projection_matrix(self):
        fx, fy = self.fx, self.fy
        cx, cy = self.cx, self.cy
        return np.array([[fx, 0, cx], [0, fy, cy], [0, 0, 1]])

    def get_projection_matrix_opengl(self):
        w, h = self.width, self.height
        fx, fy = self.fx, self.fy
        cx, cy = self.cx, self.cy
        near, far = self.near, self.far

        x_scale = 2 / w * fx
        y_scale = 2 / h * fy
        x_shift = 1 - 2 * cx / w
        y_shift = (2 * cy - h) / h
        return np.array(
            [
                [x_scale, 0, x_shift, 0],
                [0, y_scale, y_shift, 0],
                [0, 0, (near + far) / (near - far), 2 * near * far / (near - far)],
                [0, 0, -1, 0],
            ]
        ).flatten(order="F")

    def depth_to_points(self, depth, eps=0.01):
        height, width = depth.shape
        X, Y = np.meshgrid(np.arange(width), np.arange(height), indexing="xy")
        pixels = np.stack([X, Y]).reshape(2, -1)
        pixels_homo_coord = np.vstack([pixels, np.ones(pixels.shape[1])])
        P_inv = np.linalg.inv(self.get_projection_matrix())
        obj_pixels_norm = (P_inv @ pixels_homo_coord).T
        points_cam = np.einsum("ij,i->ij", obj_pixels_norm, depth.flatten())
        is_valid = (self.near + eps <= depth) & (depth <= self.far - eps)
        return points_cam[is_valid.flatten()]


@define
class Camera:
    world: World
    intrinsic: CameraIntrinsic

    @staticmethod
    def get_look_at_pose(
        eye_pos: ArrayLike, target_pos=np.zeros(3), up_vector=np.array([0.0, 0, 1])
    ):
        f = np.asarray(target_pos) - np.asarray(eye_pos)
        f /= np.linalg.norm(f)
        s = np.cross(f, up_vector)
        s /= np.linalg.norm(s)
        u = np.cross(s, f)
        rot_mat = np.vstack([s, -u, f]).T
        t = np.asarray(eye_pos)
        return SE3(SO3.from_matrix(rot_mat), t)

    def render(self, cam_pose: SE3, render_mode="tiny"):
        """output: rgb, depth, seg"""
        cam_pose_opengl = cam_pose @ SE3(SO3.from_euler("xyz", [np.pi, 0, 0]))
        view_matrix = list(cam_pose_opengl.inv().as_matrix().flatten("F"))
        proj_matrix = list(self.intrinsic.get_projection_matrix_opengl())

        # result: (width, height, rgb, depth, seg)
        renderer = (
            p.ER_BULLET_HARDWARE_OPENGL if render_mode != "tiny" else p.ER_TINY_RENDERER
        )
        result = self.world.getCameraImage(
            width=self.intrinsic.width,
            height=self.intrinsic.height,
            viewMatrix=view_matrix,
            projectionMatrix=proj_matrix,
            renderer=renderer,
        )
        w, h = self.intrinsic.width, self.intrinsic.height
        far, near = self.intrinsic.far, self.intrinsic.near
        rgb = np.reshape(result[2], (h, w, 4))[:, :, :3] * 1.0 / 255.0
        depth = np.asarray(result[3]).reshape(h, -1)
        seg = np.asarray(result[4])
        depth = far * near / (far - (far - near) * depth)
        return rgb, depth, seg


# # add noise to depth image
# def add_depth_noise(
#     depth,
#     noise_std=0.,
#     inlier_distance=0.05,
#     thres=0.5,
#     size_filt=None,
#     device="cpu"
# ):
#     import torch
#     to_tensor = lambda x, device: torch.tensor(x, device=device, dtype=torch.float32)
#     to_numpy = lambda x: x.detach().cpu().numpy()

#     def _add_gaussian_shifts(depth, std=1.):
#         depth_ = to_tensor(depth, device) # torch.tensor(depth).float().to(device)
#         depth_ = depth_.unsqueeze(0).unsqueeze(0)
#         r, c = depth.shape

#         rr = torch.arange(r, device=device)
#         cc = torch.arange(c, device=device)
#         grid_y, grid_x = torch.meshgrid(rr, cc, indexing="ij")
#         grid = torch.stack([grid_x, grid_y], dim=-1).float()
#         grid += torch.randn(r, c, 2, device=device) * std

#         grid[..., 0].clamp_(0, c-1)
#         grid[..., 1].clamp_(0, r-1)

#         n_grid = 2 * grid / torch.tensor([c-1, r-1], device=device) - 1
#         n_grid = n_grid.unsqueeze(0)
#         shifted_depth = torch.nn.functional.grid_sample(
#             depth_, n_grid,
#             mode="bilinear", padding_mode="zeros", align_corners=True)
#         return to_numpy(shifted_depth.squeeze())

#     def _simulate_depth_shade(depth, size_filt=5, window_inlier_distance=0.05, thres=0.5):
#         INVALID_VAL = 2.
#         depth_ = to_tensor(depth, device)
#         pad_size = center = size_filt // 2

#         # unfold image to work with GPU
#         depth_padded = torch.nn.functional.pad(depth_, (pad_size, pad_size, pad_size, pad_size), mode="constant", value=INVALID_VAL)
#         depth_padded = depth_padded.unsqueeze(0).unsqueeze(0)
#         unfold = torch.nn.Unfold(kernel_size=(size_filt, size_filt))
#         depth_unfolded = unfold(depth_padded)
#         depth_unfolded = depth_unfolded.squeeze(0)

#         # ignore if center pixel is outlier
#         depth_center = depth_unfolded[center*size_filt + center, :]
#         center_valid_mask = depth_center < INVALID_VAL
#         n_valids = (depth_unfolded < INVALID_VAL).sum(dim=0)
#         total_valid_mask = n_valids > (size_filt * size_filt)*thres

#         # remove if the kernel deviates a lot
#         valid_mask_elementwise = (depth_unfolded < INVALID_VAL)
#         depth_unfolded_nan = depth_unfolded.clone()
#         depth_unfolded_nan[~valid_mask_elementwise] = torch.nan
#         center = torch.nanmean(depth_unfolded_nan, dim=0)
#         diffs = torch.abs(depth_unfolded - center)
#         valids = diffs < window_inlier_distance
#         n_valids = valids.sum(dim=0).float()
#         inlier_mask = n_valids > (size_filt * size_filt)*thres

#         # update final result
#         out_depth = torch.full_like(depth_, INVALID_VAL, device=device)
#         valid_mask = center_valid_mask & total_valid_mask & inlier_mask
#         accu = depth_center[valid_mask] #torch.round(depth_center[valid_mask]*8)/8.
#         out_depth_flat = out_depth.view(-1)
#         out_depth_flat[valid_mask] = accu
#         return to_numpy(out_depth)

#     if size_filt is None:
#         size_filt = np.random.choice([3, 5, 7, 9])
#     if noise_std != 0.:
#         depth = _add_gaussian_shifts(depth, std=noise_std/2)
#     depth = _simulate_depth_shade(
#         depth, size_filt=size_filt-2,
#         window_inlier_distance=0.1,
#         thres=thres/2)

#     if noise_std != 0.:
#         depth = _add_gaussian_shifts(depth, std=noise_std)
#     depth = _simulate_depth_shade(
#         depth, size_filt=size_filt,
#         window_inlier_distance=inlier_distance,
#         thres=thres)
#     return depth



from __future__ import annotations

# from .utils import generate_temp_urdf
from pathlib import Path

import trimesh
from attrs import define
from ezpose import SE3
from numpy.typing import ArrayLike

from .shape import ShapeBase, SphereShape, CylinderShape, BoxShape, MeshShape
from .world import Body, World


@define
class GeometryBase(Body):
    shape: ShapeBase

    @classmethod
    def make_geometry_body(
        cls,
        name: str,
        world: World,
        vis_id: int,
        col_id: int,
        mass: float,
        shape: ShapeBase,
    ):
        if name in world.bodies:
            print("Body name already exists.")
            return world.bodies[name]

        uid = world.createMultiBody(
            baseVisualShapeIndex=vis_id, 
            baseCollisionShapeIndex=col_id, 
            baseMass=mass
        )
        body = cls(
            world=world, uid=uid, name=name, mass=mass, ghost=shape.ghost, shape=shape
        )
        world.bodies[name] = body
        return body


@define(repr=False)
class Sphere(GeometryBase):
    @classmethod
    def create(
        cls,
        name: str,
        world: World,
        radius: float,
        mass: float = 0.5,
        rgba: tuple[float] = (1, 0, 0, 1),
        ghost: bool = False,
    ):
        shape = SphereShape(rgba=rgba, ghost=ghost, radius=radius)
        vis_id, col_id = world.get_shape_id(shape)
        return cls.make_geometry_body(name, world, vis_id, col_id, mass, shape)


@define(repr=False)
class Cylinder(GeometryBase):
    @classmethod
    def create(
        cls,
        name: str,
        world: World,
        radius: float,
        length: float,
        mass: float = 0.5,
        rgba: tuple[float] = (1, 0, 0, 1),
        offset: SE3 = SE3.identity(),
        ghost: bool = False,
    ):
        shape = CylinderShape(
            rgba=rgba,
            ghost=ghost,
            radius=radius,
            length=length,
            offset=offset,
        )
        vis_id, col_id = world.get_shape_id(shape)
        return cls.make_geometry_body(name, world, vis_id, col_id, mass, shape)


@define(repr=False)
class Box(GeometryBase):
    @classmethod
    def create(
        cls,
        name: str,
        world: World,
        half_extents: ArrayLike,
        mass: float = 0.5,
        rgba: tuple[float] = (1, 1, 1, 1),
        ghost: bool = False,
    ):
        shape = BoxShape(rgba=rgba, ghost=ghost, half_extents=half_extents)
        vis_id, col_id = world.get_shape_id(shape)
        return cls.make_geometry_body(name, world, vis_id, col_id, mass, shape)


@define(repr=False)
class Mesh(GeometryBase):
    @classmethod
    def create(
        cls,
        name: str,
        world: World,
        visual_mesh_path: str | Path | None,
        col_mesh_path: str | Path | None = None,
        offset: SE3 = None,
        scale: float = 1.0,
        mass: float = 0.5,
        rgba: tuple[float] = (1, 1, 1, 1),
    ):
        """We assume that visual mesh is in the same coordinate as collision mesh"""
        # i think centering should be dealt with mesh level
        # centering_type:str|None = "bb", # bb(bounding box center), centroid, None
        # center = Mesh.get_center(visual_mesh_path, centering_type, scale=scale)
        # offset = SE3(trans=-center)  #in pybullet, scale is applied first
        if offset is None:
            offset = SE3()
        ghost = True if col_mesh_path is None else False
        shape = MeshShape(
            visual_mesh_path=visual_mesh_path,
            col_mesh_path=col_mesh_path,
            offset=offset,
            scale=scale,
            rgba=rgba,
            ghost=ghost,
        )
        vis_id, col_id = world.get_shape_id(shape)
        return cls.make_geometry_body(name, world, vis_id, col_id, mass, shape)

    def as_trimesh(self, is_col: bool = False):
        """if is_col is True, return the collision mesh, else return the visual mesh"""
        shape: MeshShape = self.shape
        mesh_path = shape.visual_mesh_path

        if is_col:
            assert shape.col_mesh_path is not None, "No collision mesh"
            mesh_path = shape.col_mesh_path
        return (
            trimesh.load(mesh_path)
            .apply_translation(shape.offset.trans / shape.scale)
            .apply_scale(shape.scale)
        )

    @classmethod
    def from_trimesh(
        cls,
        name,
        world: World,
        mesh: trimesh.Trimesh,
        col_mesh: trimesh.Trimesh | None = None,
        mass: float = 0.5,
        rgba: ArrayLike = [1, 1, 1, 1],
        offset: SE3 = None,
        # centering_type=None,
    ):
        if offset is None:
            offset = SE3()
        import tempfile

        tempdir = tempfile.TemporaryDirectory()
        if col_mesh is None:
            col_mesh = mesh

        mesh_path = Path(tempdir.name) / "mesh.obj"
        col_mesh_path = Path(tempdir.name) / "col_mesh.obj"
        mesh.export(mesh_path, "obj")
        col_mesh.export(col_mesh_path, "obj")
        obj = cls.create(
            name,
            world,
            str(mesh_path),
            str(col_mesh_path),
            offset=offset,
            # centering_type=centering_type,
            mass=mass,
            rgba=rgba,
        )
        tempdir.cleanup()
        return obj
